--------------------------------
-------Resource Aware ML--------
--------------------------------
-- File:
--   CallGraph.hs
--
-- Author:
--   Jan Hoffmann (2009)
--
-- Description: 
--   Stuff that deals with the call graph of a program.
--
--   The function callGraph takes a list of function definitions and
--   coumputes the call graph.
--
--   The function getSCCs computes the strongly connected components
--   of the call graph from a list of function definitions..

module CallGraph where

import Data.List
import Data.Graph

import RAMLTypes

-- Following the GHC graph library a graph (in list form) is a list of
-- triples (nodeName, nodeKey, successorKeys).

callGraph funList = [ (fun,f,functionCalls e_f) | fun@(f,_,e_f,_) <- funList ]


functionCalls expr = 
    case expr of
      ETick _ _ -> []
      EUnit _ -> []
      ETrue _ -> []
      EFalse _ -> []
      ENum _ _ -> []
      EVar _ _ -> []

      EApp _ fName e _ -> [fName] `union` functionCalls e
      EUnaryOp _ e _ -> functionCalls e
      EBinOp _ e1 e2 _ -> (functionCalls e1) `union` (functionCalls e2)

      ESucc e _ -> functionCalls e
      EToInt e _ -> functionCalls e
      EMatchN e e1 _ e2 _ -> foldl union [] [functionCalls e | e <- [e,e1,e2] ]

      ELet x e1 e2 _ -> (functionCalls e1) `union` (functionCalls e2)
      ECond e e1 e2 _ -> foldl union [] [functionCalls e | e <- [e,e1,e2] ]

      ENil _ -> []
      ECons e1 e2 _ ->  (functionCalls e1) `union` (functionCalls e2)
      EMatchL _ e e1 x1 x2 e2 _ -> foldl union [] [functionCalls e | e <- [e,e1,e2] ]

      ETuple es _ -> foldl union [] [functionCalls e | e <- es]
      EMatchT e xs e' _ -> (functionCalls e) `union` (functionCalls e')

      ELeaf _ -> []
      ENodeB e1 e2 e3 _ -> foldl union [] [functionCalls e | e <- [e1,e2,e3] ]
      EMatchB _ e e1 _ _ _ e2 _ -> foldl union [] [functionCalls e | e <- [e,e1,e2] ]



getSCCs funList = map flattenSCC (stronglyConnComp (callGraph funList) ) 
-- to be in reverse topological order.   C -> G means that G is before C
-- in the list of SCCs.

