--------------------------------
-------Resource Aware ML--------
--------------------------------
-- File:
--   Clone.hs
--
-- Author:
--   Klaus Aehlig (2011)
--
-- Description:
--    For every type declare a function that clones an element of that type

module Clone where

import RAMLTypes

import List
import Text.ParserCombinators.Parsec (SourcePos)

subtypes :: Type -> [Type]
-- for a  given type, return the list of all its subtypes (including itself)
-- in increasing order

subtypes TInt       = [TInt]
subtypes TBool      = [TBool]
subtypes (TList a)  = (subtypes a) ++ [TList a]
subtypes (TBTree a) = (subtypes a) ++ [TBTree a]
subtypes (TTuple l) = (foldl union [] (map subtypes l)) ++ [TTuple l]

typename :: Type -> String
-- for a given type, return a unique string representation that can be used as variable name

typename TInt       = "i"
typename TBool      = "b"
typename (TList a)  = "l" ++ (typename a)
typename (TBTree a) = "b" ++ (typename a)
typename (TTuple l) = "t" ++ (show (length l)) ++ (foldl (++) "" (map typename l))

clonename :: Type -> String
-- for a given type, return the name of the corresponing clone function

clonename t = "CLONE_" ++ (typename t) 

clonevar t     = TEVar "x" t
clonevarx t x  = TEVar x   t

clone :: Type -> TExp
-- function body of the clone routine of the corresponding type

clone TInt    = TETuple Normal [clonevar TInt, clonevar TInt] (TTuple [TInt, TInt])
clone TBool   = TETuple Normal [clonevar TBool, clonevar TBool] (TTuple [TBool, TBool])
clone (TList a) = TEMatchL Dest (clonevar (TList a)) (TENil Normal (TList a))  "h" "t"
                  (TEMatchT Normal (TEApp Nothing (clonename a) (clonevarx a "h") (TTuple [a,a])) ["h1", "h2"]
                   (TEMatchT Normal (TEApp Nothing (clonename (TList a)) (clonevarx (TList a) "t") (TTuple [TList a, TList a])) ["t1", "t2"]
                    (TETuple Normal [TECons Normal (clonevarx a "h1") (clonevarx (TList a) "t1") (TList a), 
                              TECons Normal (clonevarx a "h2") (clonevarx (TList a) "t2") (TList a)] 
                     (TTuple [TList a, TList a])) (TTuple [TList a, TList a])) (TTuple [TList a, TList a])) (TTuple [TList a, TList a])

-- clone (TBTree a) = EMatchB Dest clonevar (ELeaf clonesource)  "h" "tl" "tr"
--                   (EMatchT (EApp (clonename a) (clonevarx "h") clonesource) ["h1", "h2"]
--                    (EMatchT (EApp (clonename (TBTree a)) (clonevarx "tl") clonesource) ["tl1", "tl2"]
--                     (EMatchT (EApp (clonename (TBTree a)) (clonevarx "tr") clonesource) ["tr1", "tr2"]
--                      (ETuple [ENodeB (clonevarx "h1") (clonevarx "tl1")(clonevarx "tr1") clonesource, 
--                               ENodeB (clonevarx "h2") (clonevarx "tl2")(clonevarx "tr2") clonesource] 
--                       clonesource) clonesource) clonesource) clonesource) clonesource
